use std::ffi::{c_void, CStr};
use windows::core::PCSTR;
use windows::Win32::Foundation::{HANDLE, HMODULE};
use windows::Win32::System::LibraryLoader::{GetModuleFileNameA, GetModuleHandleA, GetProcAddress};
use windows::Win32::System::ProcessStatus::{EnumProcessModules, GetModuleInformation, MODULEINFO};
use windows::Win32::System::Threading::GetCurrentProcess;

pub type FarProcUnwrapped = unsafe extern "system" fn() -> isize;

#[derive(thiserror::Error, Debug)]
pub enum Error {
    #[error("Windows Error: {0}")]
    Windows(#[from] windows::core::Error),
    #[error("FromUtf8 Error: {0}")]
    FromUtf8(#[from] std::string::FromUtf8Error),
    #[error("Symbol {0} not found")]
    SymbolNotFound(String),
    #[error("No module found for handle")]
    ModuleNotFound,
}

pub struct ModuleInfo {
    pub base: usize,
    pub end: usize,
    pub name: String,
}

pub fn get_module_base(module_name: Option<PCSTR>) -> HMODULE {
    match module_name {
        None => {
            // Infallible, safe unwrap since GetModuleHandleA(null) has to always return
            let module = unsafe { GetModuleHandleA(PCSTR::null()) }.unwrap();
            println!("Main executable base: {:X}", module.0 as usize);
            module
        }
        Some(name) => {
            let module = loop {
                match unsafe { GetModuleHandleA(name) } {
                    Ok(handle) => break handle,
                    Err(_) => std::thread::sleep(std::time::Duration::from_millis(1)),
                }
            };
            unsafe {
                println!(
                    "{} base: {:X}",
                    name.to_string().unwrap(), // Name should be a valid string
                    module.0 as usize
                );
            }
            module
        }
    }
}

pub fn get_symbol(module: usize, func_name: PCSTR) -> Result<FarProcUnwrapped, Error> {
    let func_name_str = unsafe { func_name.to_string() }?;
    match unsafe { GetProcAddress(HMODULE(module as *mut c_void), func_name) } {
        None => Err(Error::SymbolNotFound(func_name_str)),
        Some(ptr) => {
            println!("Symbol {func_name_str} found at 0x{:X}", ptr as usize);
            Ok(ptr)
        }
    }
}

pub fn get_symbol_ptr(module: usize, func_name: PCSTR) -> Result<usize, Error> {
    Ok(unsafe { std::mem::transmute::<FarProcUnwrapped, usize>(get_symbol(module, func_name)?) })
}

pub fn get_current_process() -> HANDLE {
    unsafe { GetCurrentProcess() }
}

pub fn find_module_for_address(address: usize, debug: bool) {
    let process = get_current_process();

    match enumerate_modules(process, debug) {
        Err(err) => println!("Could not find module for address {address}, error: {err}"),
        Ok(modules) => {
            find_module_for_address_simple(address, process, modules);
        }
    }
}

pub fn find_module_for_address_simple(address: usize, process: HANDLE, modules: Vec<HMODULE>) {
    for &module in &modules {
        match get_module_info(process, module) {
            Err(err) => println!("Could not find module for address {address}, error: {err}"),
            Ok(info) => {
                if address >= info.base && address < info.end {
                    println!(
                        "Address {:p} belongs to: {} with base: {:p} so offset is: {:p}",
                        address as *const u8,
                        info.name,
                        info.base as *const u8,
                        (address - info.base) as *const u8
                    );
                }
            }
        }
    }
}

pub fn enumerate_modules(process: HANDLE, debug: bool) -> Result<Vec<HMODULE>, Error> {
    let mut modules: [HMODULE; 1024] = [HMODULE(std::ptr::null_mut()); 1024];
    let mut cb_needed = 0;

    unsafe {
        EnumProcessModules(
            process,
            modules.as_mut_ptr(),
            size_of_val(&modules) as u32,
            &mut cb_needed,
        )
    }
    .map_err(Error::Windows)?;

    let module_count = cb_needed as usize / size_of::<HMODULE>();
    let sub_slice = &modules[..module_count];
    if debug {
        for &module in sub_slice {
            if let Ok(info) = get_module_info(process, module) {
                println!(
                    "Loaded module: {}, with mapped range {:p}-{:p}",
                    info.name, info.base as *const u8, info.end as *const u8,
                );
            }
        }
    }
    Ok(sub_slice.to_vec())
}

pub fn get_module_info(process: HANDLE, module: HMODULE) -> Result<ModuleInfo, Error> {
    let mut mod_info = MODULEINFO::default();

    unsafe {
        GetModuleInformation(
            process,
            module,
            &mut mod_info,
            size_of::<MODULEINFO>() as u32,
        )
    }
    .map_err(Error::Windows)?;

    let base = mod_info.lpBaseOfDll as usize;
    let end = base + mod_info.SizeOfImage as usize;
    let name = get_module_filename(module)?;

    Ok(ModuleInfo { base, end, name })
}

fn get_module_filename(module: HMODULE) -> Result<String, Error> {
    let mut buffer = [0u8; 512];
    let len = unsafe { GetModuleFileNameA(Some(module), &mut buffer) };
    if len > 0 {
        let cstr = unsafe { CStr::from_ptr(buffer.as_ptr() as *const i8) };
        Ok(cstr.to_string_lossy().into_owned())
    } else {
        Err(Error::ModuleNotFound)
    }
}
